/*
 *
 * X DLX Simulator
 *
 * The DLX is a 32-bit RISC machine based on the MIPS 2000 architecture.
 * The architecture includes 32 32-bit general purpose registers and a 
 * program counter.
 *
 * Floating point registers and an interrupt register are not implemented.
 *
 * The address space is 4 Gbytes, big-endian byte-addressable using 32-bit
 * addresses. All instructions must be properly aligned.
 *
 *
 * Cameron Humphries
 * Department of Computer Science
 * University of Adelaide
 * 1992
 *
 */
 
#include <stdio.h>
#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Dialog.h>

#include "sim.h"

extern Widget toplevel,form;
extern XtAppContext app_context;

static String reg_strings[] = {
  "r0_label"," r0","r0_value","0000000000","r0_type","DEC",
  "r1_label"," r1","r1_value","0000000000","r1_type","DEC",
  "r2_label"," r2","r2_value","0000000000","r2_type","DEC",
  "r3_label"," r3","r3_value","0000000000","r3_type","DEC",
  "r4_label"," r4","r4_value","0000000000","r4_type","DEC",
  "r5_label"," r5","r5_value","0000000000","r5_type","DEC",
  "r6_label"," r6","r6_value","0000000000","r6_type","DEC",
  "r7_label"," r7","r7_value","0000000000","r7_type","DEC",
  "r8_label"," r8","r8_value","0000000000","r8_type","DEC",
  "r9_label"," r9","r9_value","0000000000","r9_type","DEC",
  "r10_label","r10","r10_value","0000000000","r10_type","DEC",
  "r11_label","r11","r11_value","0000000000","r11_type","DEC",
  "r12_label","r12","r12_value","0000000000","r12_type","DEC",
  "r13_label","r13","r13_value","0000000000","r13_type","DEC",
  "r14_label","r14","r14_value","0000000000","r14_type","DEC",
  "r15_label","r15","r15_value","0000000000","r15_type","DEC",
  "r16_label","r16","r16_value","0000000000","r16_type","DEC",
  "r17_label","r17","r17_value","0000000000","r17_type","DEC",
  "r18_label","r18","r18_value","0000000000","r18_type","DEC",
  "r19_label","r19","r19_value","0000000000","r19_type","DEC",
  "r20_label","r20","r20_value","0000000000","r20_type","DEC",
  "r21_label","r21","r21_value","0000000000","r21_type","DEC",
  "r22_label","r22","r22_value","0000000000","r22_type","DEC",
  "r23_label","r23","r23_value","0000000000","r23_type","DEC",
  "r24_label","r24","r24_value","0000000000","r24_type","DEC",
  "r25_label","r25","r25_value","0000000000","r25_type","DEC",
  "r26_label","r26","r26_value","0000000000","r26_type","DEC",
  "r27_label","r27","r27_value","0000000000","r27_type","DEC",
  "r28_label","r28","r28_value","0000000000","r28_type","DEC",
  "r29_label","r29","r29_value","0000000000","r29_type","DEC",
  "r30_label","r30","r30_value","0000000000","r30_type","DEC",
  "r31_label","r31","r31_value","0000000000","r31_type","DEC"
  };  

#define SHOW_SIGNED 1
#define SHOW_OCTAL 2
#define SHOW_HEX 3
#define SHOW_UNSIGNED 4
#define SHOW_ASCII 5

char *regs_buf[33];
int regs_type[33];

static Widget reg_widg = NULL;
static Widget edit_widg = NULL;
static Widget dialog_widg;
  
static void sign_decimal_val(),unsign_decimal_val();
static void octal_val();
static void hex_val(),ascii_val();
static void edit_register_value();
static void FindWidget(),EditWidget();
static void dialog_ok(),dialog_cancel();

/*
 * 
 */
void Create_Register_View(Widget form)
{
  Widget w,popup_w;
  static Arg label_args[] = {
    {XtNlabel, (XtArgVal)NULL},
    {XtNborderWidth, (XtArgVal)1},
    {XtNfromHoriz, (XtArgVal)NULL},
    {XtNfromVert, (XtArgVal)NULL},
    {XtNhorizDistance, (XtArgVal)5},
  };
  static Arg menu_entry_args[] = {
    {XtNlabel, (XtArgVal)NULL},
    {XtNcallback, (XtArgVal)NULL},
  };
  static XtCallbackRec unsign_decimal_cbk[] = {
    {unsign_decimal_val, NULL},
    {NULL, NULL},
  };
  static XtCallbackRec octal_cbk[] = {
    {octal_val, NULL},
    {NULL, NULL},
  };
  static XtCallbackRec sign_decimal_cbk[] = {
    {sign_decimal_val, NULL},
    {NULL, NULL},
  };
  static XtCallbackRec hex_cbk[] = {
    {hex_val, NULL},
    {NULL, NULL},
  };
  static XtCallbackRec ascii_cbk[] = {
    {ascii_val, NULL},
    {NULL, NULL},
  };
  static Arg dialog_args[] = {
    {XtNlabel,(XtArgVal)"Edit register value:"},
    {XtNvalue, (XtArgVal)""},
  };
  
  static String type_trans = 
    "<Btn3Down>:FindWidget() XawPositionSimpleMenu(reg_menu) MenuPopup(reg_menu)";
  static String PC_type_trans = 
    "<Btn3Down>:FindWidget() XawPositionSimpleMenu(pc_menu) MenuPopup(pc_menu)";
  static String edit_trans = 
    "<Btn1Down>(2):EditWidget() edit_reg()"; 
  static String reg_dialog_trans = 
    "<Key>Return: RegDialogOK() \n\
     <Key>KP_Enter: RegDialogOK()";
    
  static XtActionsRec act_table[] = {
    {"edit_reg",edit_register_value},
    {"FindWidget",FindWidget},
    {"EditWidget",EditWidget},
    {"RegDialogOK",dialog_ok},
  };
    
  XtAppAddActions(app_context,act_table,XtNumber(act_table));
  
  /* Create array of strings holding register values */
  {
    int i;
    
    for (i=0;i<33;i++)
      {
	regs_buf[i] = malloc(15);
	if (i==32)
	  {
	    regs_type[i] = SHOW_UNSIGNED;
	    sprintf(regs_buf[i],"%12u",regs[i]);
	  }
	else
	  {
	    regs_type[i] = SHOW_SIGNED;
	    sprintf(regs_buf[i],"%12d",regs[i]);
	  }
      }
  }

  /* Create popup representation menu */
  popup_w = XtCreatePopupShell("reg_menu",simpleMenuWidgetClass,form,NULL,0);
  
  menu_entry_args[0].value = (XtArgVal)"OCTAL";
  menu_entry_args[1].value = (XtArgVal)octal_cbk;
  w = XtCreateManagedWidget("octal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"SIGNED DECIMAL";
  menu_entry_args[1].value = (XtArgVal)sign_decimal_cbk;
  w = XtCreateManagedWidget("sign_decimal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"UNSIGNED DECIMAL";
  menu_entry_args[1].value = (XtArgVal)unsign_decimal_cbk;
  w = XtCreateManagedWidget("unsign_decimal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"HEX";
  menu_entry_args[1].value = (XtArgVal)hex_cbk;
  w = XtCreateManagedWidget("hex",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"ASCII";
  menu_entry_args[1].value = (XtArgVal)ascii_cbk;
  w = XtCreateManagedWidget("ascii",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));

  /* Create popup representation menu */
  popup_w = XtCreatePopupShell("pc_menu",simpleMenuWidgetClass,form,NULL,0);
  
  menu_entry_args[0].value = (XtArgVal)"OCTAL";
  menu_entry_args[1].value = (XtArgVal)octal_cbk;
  w = XtCreateManagedWidget("octal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"UNSIGNED DECIMAL";
  menu_entry_args[1].value = (XtArgVal)unsign_decimal_cbk;
  w = XtCreateManagedWidget("unsign_decimal",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  menu_entry_args[0].value = (XtArgVal)"HEX";
  menu_entry_args[1].value = (XtArgVal)hex_cbk;
  w = XtCreateManagedWidget("hex",smeBSBObjectClass,popup_w,
			    menu_entry_args,XtNumber(menu_entry_args));
			    
  /* Create popup dialog box for editing values */
  w = XtCreatePopupShell("reg_edit",overrideShellWidgetClass,form,NULL,0);
  dialog_widg = XtCreateManagedWidget("reg_dialog",dialogWidgetClass,w,
			    dialog_args,XtNumber(dialog_args));
  XawDialogAddButton(dialog_widg,"OK",dialog_ok,NULL);
  XawDialogAddButton(dialog_widg,"Cancel",dialog_cancel,NULL);
  XtOverrideTranslations(XtNameToWidget(dialog_widg,"value"),
			 XtParseTranslationTable(reg_dialog_trans));
  
  /* Create PC view */
  label_args[0].value = (XtArgVal)" PC";
  label_args[1].value = (XtArgVal)1;
  label_args[2].value = (XtArgVal)XtNameToWidget(form,"Code_View");
  label_args[3].value = (XtArgVal)XtNameToWidget(form,"Execute_Menu");
  w = XtCreateManagedWidget("PC_label",labelWidgetClass,form,
			    label_args,XtNumber(label_args));
			    
  label_args[0].value = (XtArgVal)regs_buf[32];
  label_args[1].value = (XtArgVal)1;
  label_args[2].value = (XtArgVal)w;
  label_args[3].value = (XtArgVal)XtNameToWidget(form,"Execute_Menu");
  w = XtCreateManagedWidget("PC_value",labelWidgetClass,form,
			    label_args,XtNumber(label_args));
  XtOverrideTranslations(w,XtParseTranslationTable(edit_trans));
			    
  label_args[0].value = (XtArgVal)"UNS";
  label_args[1].value = (XtArgVal)1;
  label_args[2].value = (XtArgVal)w;
  label_args[3].value = (XtArgVal)XtNameToWidget(form,"Execute_Menu");
  w = XtCreateManagedWidget("PC_type",labelWidgetClass,form,
			    label_args,XtNumber(label_args));
  XtOverrideTranslations(w,XtParseTranslationTable(PC_type_trans));
			    
  /* Create 32 Register views */
  {
    int i=0,j=0;
    for (i=0;i<16;i++) {
      label_args[0].value = (XtArgVal)reg_strings[j+1];
      label_args[1].value = (XtArgVal)1;
      label_args[2].value = (XtArgVal)XtNameToWidget(form,"Code_View");
      if (j) {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-6]);
      }
      else {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,"PC_label");
      }      
      w = XtCreateManagedWidget(reg_strings[j],labelWidgetClass,form,
				label_args,XtNumber(label_args));
			    
      label_args[0].value = (XtArgVal)regs_buf[i];
      label_args[1].value = (XtArgVal)1;
      label_args[2].value = (XtArgVal)w;
      if (j) {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-4]);
      }
      else {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,"PC_label");
      }      
      w = XtCreateManagedWidget(reg_strings[j+2],labelWidgetClass,form,
				label_args,XtNumber(label_args));
      if (i)
	XtOverrideTranslations(w,XtParseTranslationTable(edit_trans));
			    
      label_args[0].value = (XtArgVal)reg_strings[j+5];
      label_args[1].value = (XtArgVal)1;
      label_args[2].value = (XtArgVal)w;
      if (j) {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-2]);
      }
      else {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,"PC_label");
      }      
      w = XtCreateManagedWidget(reg_strings[j+4],labelWidgetClass,form,
				label_args,XtNumber(label_args));
      XtOverrideTranslations(w,XtParseTranslationTable(type_trans));
      j += 6;
    }    

    for (i=16;i<32;i++) {
      label_args[0].value = (XtArgVal)reg_strings[j+1];
      label_args[1].value = (XtArgVal)1;
      label_args[2].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-(16*6)+4]);
      if (j != (16*6)) {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-6]);
      }
      else {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,"PC_label");
      }      
      label_args[4].value = (XtArgVal)10;
      w = XtCreateManagedWidget(reg_strings[j],labelWidgetClass,form,
				label_args,XtNumber(label_args));
			    
      label_args[0].value = (XtArgVal)regs_buf[i];
      label_args[1].value = (XtArgVal)1;
      label_args[2].value = (XtArgVal)w;
      if (j != (16*6)) {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-4]);
      }
      else {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,"PC_label");
      }      
      label_args[4].value = (XtArgVal)5;
      w = XtCreateManagedWidget(reg_strings[j+2],labelWidgetClass,form,
				label_args,XtNumber(label_args));
      XtOverrideTranslations(w,XtParseTranslationTable(edit_trans));
			    
      label_args[0].value = (XtArgVal)reg_strings[j+5];
      label_args[1].value = (XtArgVal)1;
      label_args[2].value = (XtArgVal)w;
      if (j != (16*6)) {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,reg_strings[j-2]);
      }
      else {
	label_args[3].value = (XtArgVal)XtNameToWidget(form,"PC_label");
      }      
      w = XtCreateManagedWidget(reg_strings[j+4],labelWidgetClass,form,
				label_args,XtNumber(label_args));
      XtOverrideTranslations(w,XtParseTranslationTable(type_trans));
			    
      j += 6;
    }    
  }

  return;
}

static void unsign_decimal_val(Widget w,XtPointer client_data, XtPointer call_data)
{
  int index;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  Widget reg;
  char buf[20];
  
  if (strcmp(XtName(reg_widg),"PC_type"))
    {
      sscanf(XtName(reg_widg),"%*c%d%*s",&index);
      regs_type[index] = SHOW_UNSIGNED;
      sprintf(regs_buf[index],"%12u",regs[index]);
  
      sprintf(buf,"r%1d_value",index);
      reg = XtNameToWidget(form,buf);
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
  
      label_arg[0].value = (XtArgVal)"UNS";
      XtSetValues(reg_widg,label_arg,1);
    }
  else
    {
      index = 32;
      regs_type[index] = SHOW_UNSIGNED;
      sprintf(regs_buf[index],"%12u",regs[index]);
      reg = XtNameToWidget(form,"PC_value");
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
  
      label_arg[0].value = (XtArgVal)"UNS";
      XtSetValues(reg_widg,label_arg,1);
    }
    
  return;
}

static void octal_val(Widget w,XtPointer client_data, XtPointer call_data)
{
  int index;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  Widget reg;
  char buf[20];
  
  if (strcmp(XtName(reg_widg),"PC_type"))
    {
      sscanf(XtName(reg_widg),"%*c%d%*s",&index);
      regs_type[index] = SHOW_OCTAL;
      sprintf(regs_buf[index],"%12o",regs[index]);
  
      sprintf(buf,"r%1d_value",index);
      reg = XtNameToWidget(form,buf);
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
  
      label_arg[0].value = (XtArgVal)"OCT";
      XtSetValues(reg_widg,label_arg,1);
    }
  else
    {
      index = 32;
      regs_type[index] = SHOW_OCTAL;
      sprintf(regs_buf[index],"%12o",regs[index]);
      reg = XtNameToWidget(form,"PC_value");
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
  
      label_arg[0].value = (XtArgVal)"OCT";
      XtSetValues(reg_widg,label_arg,1);
    }
    
  return;
}

static void sign_decimal_val(Widget w,XtPointer client_data, XtPointer call_data)
{
  int index;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  Widget reg;
  char buf[20];
  
  sscanf(XtName(reg_widg),"%*c%d%*s",&index);
  regs_type[index] = SHOW_SIGNED;
  sprintf(regs_buf[index],"%12d",regs[index]);
  
  sprintf(buf,"r%1d_value",index);
  reg = XtNameToWidget(form,buf);
  label_arg[0].value = (XtArgVal)regs_buf[index];
  XtSetValues(reg,label_arg,1);
  
  label_arg[0].value = (XtArgVal)"DEC";
  XtSetValues(reg_widg,label_arg,1);
  
  return;
}

static void hex_val(Widget w,XtPointer client_data, XtPointer call_data)
{
  int index;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  Widget reg;
  char buf[20];
  
  if (strcmp(XtName(reg_widg),"PC_type"))
    {
      sscanf(XtName(reg_widg),"%*c%d%*s",&index);
      regs_type[index] = SHOW_HEX;
      sprintf(regs_buf[index],"%12x",regs[index]);
  
      sprintf(buf,"r%1d_value",index);
      reg = XtNameToWidget(form,buf);
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
  
      label_arg[0].value = (XtArgVal)"HEX";
      XtSetValues(reg_widg,label_arg,1);
    }
  else
    {
      index = 32;
      regs_type[index] = SHOW_HEX;
      sprintf(regs_buf[index],"%12x",regs[index]);
      reg = XtNameToWidget(form,"PC_value");
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
  
      label_arg[0].value = (XtArgVal)"HEX";
      XtSetValues(reg_widg,label_arg,1);
    }
  
  return;
}

static void ascii_val(Widget w,XtPointer client_data, XtPointer call_data)
{
  int index;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  Widget reg;
  char buf[20];
  char *ptr;
  
  sscanf(XtName(reg_widg),"%*c%d%*s",&index);
  regs_type[index] = SHOW_ASCII;
  ptr = &regs[index];
  strcpy(regs_buf[index],"        ");
  if (ptr[0] < 32)
    strcat(regs_buf[index]," ");
  else
    {
      regs_buf[index][8] = ptr[0];
      regs_buf[index][9] = '\0';
    }
    
  if (ptr[1] < 32)
    strcat(regs_buf[index]," ");
  else
    {
      regs_buf[index][9] = ptr[1];
      regs_buf[index][10] = '\0';
    }
    
  if (ptr[2] < 32)
    strcat(regs_buf[index]," ");
  else
    {
      regs_buf[index][10] = ptr[2];
      regs_buf[index][11] = '\0';
    }
    
  if (ptr[3] < 32)
    strcat(regs_buf[index]," ");
  else
    {
      regs_buf[index][11] = ptr[3];
      regs_buf[index][12] = '\0';
    }
  
  sprintf(buf,"r%1d_value",index);
  reg = XtNameToWidget(form,buf);
  label_arg[0].value = (XtArgVal)regs_buf[index];
  XtSetValues(reg,label_arg,1);
  
  label_arg[0].value = (XtArgVal)"ASC";
  XtSetValues(reg_widg,label_arg,1);
  
  return;
}

static void edit_register_value(Widget w,XEvent *xev,String *params,Cardinal *numparams)
{
  Widget widg = XtNameToWidget(form,"reg_edit");
  Window junk1,junk2;
  int root_x,root_y,junkX,junkY;
  unsigned int junkM;
  
  if (widg != NULL)
    {
      XtRealizeWidget(widg);
      XQueryPointer(XtDisplay(widg),XtWindow(widg), &junk1, &junk2,
		    &root_x,&root_y,&junkX,&junkY,&junkM);

      XtMoveWidget(widg,root_x-40,root_y-35);
      XtPopup(widg,XtGrabExclusive);
    }
  else
    fprintf(stderr,"Oh shit!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  return;
}

static void FindWidget(Widget w,XEvent *xev,String *params,Cardinal *numparams)
{
  reg_widg = XtWindowToWidget(XtDisplay(w),xev->xbutton.window);
  
  return;
}

static void EditWidget(Widget w,XEvent *xev,String *params,Cardinal *numparams)
{
  int index,i;
  char *buf = (char *)malloc(20);
  char *ptr;
  static Arg args[] = {
    {XtNvalue, (XtArgVal)NULL},
  };
  
  edit_widg = XtWindowToWidget(XtDisplay(w),xev->xbutton.window);
  
  if (strcmp(XtName(edit_widg),"PC_value"))
    sscanf(XtName(edit_widg),"%*c%d%*s",&index);
  else
    index = 32;
    
  switch (regs_type[index])
    {
    case SHOW_OCTAL:
      sprintf(buf,"%o",regs[index]);
      break;
    case SHOW_HEX:
      sprintf(buf,"%x",regs[index]);
      break;
    case SHOW_SIGNED:
      sprintf(buf,"%d",regs[index]);
      break;
    case SHOW_UNSIGNED:
      sprintf(buf,"%u",regs[index]);
      break;
    case SHOW_ASCII:
      for (i=8;i<12;i++)
	buf[i-8] = regs_buf[index][i];
      
      break;
    }
    
  args[0].value = (XtArgVal)buf;
  XtSetValues(dialog_widg,args,1);

  return;
}

static void dialog_ok(Widget w,XEvent *xev,String *params,Cardinal *numparams)
{
  int index;
  char *buf = (char *)malloc(20);
  char *str,*ptr;
  unsigned int newval;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  Widget reg;
  
  str = XawDialogGetValueString(dialog_widg);

  if (strcmp(XtName(edit_widg),"PC_value"))
    sscanf(XtName(edit_widg),"%*c%d%*s",&index);
  else
    index=32;
  
  switch (regs_type[index])
    {
    case SHOW_OCTAL:
      sscanf(str,"%o",&newval);
      break;
    case SHOW_HEX:
      sscanf(str,"%x",&newval);
      break;
    case SHOW_SIGNED:
      sscanf(str,"%d",&newval);
      break;
    case SHOW_UNSIGNED:
      sscanf(str,"%u",&newval);
      break;
    case SHOW_ASCII:
      ptr = &newval;
      switch (strlen(str))
	{
	case 0:
	  ptr[0] = ptr[1] = ptr[2] = ptr[3] = '\0';
	  break;
	case 1:
	  ptr[0] = str[0];
	  ptr[1] = ptr[2] = ptr[3] = '\0';
	  break;
	case 2:
	  ptr[0] = str[0];
	  ptr[1] = str[1];
	  ptr[2] = ptr[3] = '\0';
	  break;
	case 3:
	  ptr[0] = str[0];
	  ptr[1] = str[1];
	  ptr[2] = str[2];
	  ptr[3] = '\0';
	  break;
	default:
	  ptr[0] = str[0];
	  ptr[1] = str[1];
	  ptr[2] = str[2];
	  ptr[3] = str[3];
	  break;
	}	
      break;
    }
  regs[index] = newval;
  
  switch (regs_type[index])
    {
    case SHOW_OCTAL:
      sprintf(regs_buf[index],"%12o",regs[index]);
      break;
    case SHOW_HEX:
      sprintf(regs_buf[index],"%12x",regs[index]);
      break;
    case SHOW_SIGNED:
      sprintf(regs_buf[index],"%12d",regs[index]);
      break;
    case SHOW_UNSIGNED:
      sprintf(regs_buf[index],"%12u",regs[index]);
      break;
    case SHOW_ASCII:
      ptr = &regs[index];
      strcpy(regs_buf[index],"        ");
      if (ptr[0] < 32)
	strcat(regs_buf[index]," ");
      else
	{
	  regs_buf[index][8] = ptr[0];
	  regs_buf[index][9] = '\0';
	}
      
      if (ptr[1] < 32)
	strcat(regs_buf[index]," ");
      else
	{
	  regs_buf[index][9] = ptr[1];
	  regs_buf[index][10] = '\0';
	}
      
      if (ptr[2] < 32)
	strcat(regs_buf[index]," ");
      else
	{
	  regs_buf[index][10] = ptr[2];
	  regs_buf[index][11] = '\0';
	}
      
      if (ptr[3] < 32)
	strcat(regs_buf[index]," ");
      else
	{
	  regs_buf[index][11] = ptr[3];
	  regs_buf[index][12] = '\0';
	}
      
      break;
    }
    
  if (index==32)
    {
      reg = XtNameToWidget(form,"PC_value");
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
      Update_Code();
    }
  else
    {
      sprintf(buf,"r%1d_value",index);
      reg = XtNameToWidget(form,buf);
      label_arg[0].value = (XtArgVal)regs_buf[index];
      XtSetValues(reg,label_arg,1);
    }
  
  XtPopdown(XtParent(XtParent(w)));
  
  return;
}

static void dialog_cancel(Widget w,XEvent *xev,String *params,Cardinal *numparams)
{
  XtPopdown(XtParent(XtParent(w)));
  
  return;
}

Update_Regs()
{
  int i=0;
  char *ptr;
  Widget reg;
  static Arg label_arg[] = {
    {XtNlabel, (XtArgVal)NULL},
  };
  char *buf = (char *)malloc(20);
      
  for (i=0;i<33;i++)
    {
      switch (regs_type[i])
	{
	case SHOW_OCTAL:
	  sprintf(regs_buf[i],"%12o",regs[i]);
	  break;
	case SHOW_HEX:
	  sprintf(regs_buf[i],"%12x",regs[i]);
	  break;
	case SHOW_SIGNED:
	  sprintf(regs_buf[i],"%12d",regs[i]);
	  break;
	case SHOW_UNSIGNED:
	  sprintf(regs_buf[i],"%12u",regs[i]);
	  break;
	case SHOW_ASCII:
	  ptr = &regs[i];
	  strcpy(regs_buf[i],"        ");
	  if (ptr[0] < 32)
	    strcat(regs_buf[i]," ");
	  else
	    {
	      regs_buf[i][8] = ptr[0];
	      regs_buf[i][9] = '\0';
	    }
	  
	  if (ptr[1] < 32)
	    strcat(regs_buf[i]," ");
	  else
	    {
	      regs_buf[i][9] = ptr[1];
	      regs_buf[i][10] = '\0';
	    }
	  
	  if (ptr[2] < 32)
	    strcat(regs_buf[i]," ");
	  else
	    {
	      regs_buf[i][10] = ptr[2];
	      regs_buf[i][11] = '\0';
	    }
	  
	  if (ptr[3] < 32)
	    strcat(regs_buf[i]," ");
	  else
	    {
	      regs_buf[i][11] = ptr[3];
	      regs_buf[i][12] = '\0';
	    }
  
	  break;
	}

      if (i==32)
	{
	  reg = XtNameToWidget(form,"PC_value");
	  label_arg[0].value = (XtArgVal)regs_buf[i];
	  XtSetValues(reg,label_arg,1);
	}
      else
	{
	  sprintf(buf,"r%1d_value",i);
	  reg = XtNameToWidget(form,buf);
	  label_arg[0].value = (XtArgVal)regs_buf[i];
	  XtSetValues(reg,label_arg,1);
	}
    }

  return;
}
